package matrices;

import java.beans.ConstructorProperties;
import java.io.*;
import java.util.ArrayList;
import java.util.StringTokenizer;

public class Matrix implements Cloneable {
    private double[][] array;

    public Matrix() {
        array = new double[0][0];
    }

    public Matrix(int m, int n) {
        array = new double[m][n];
    }

    @ConstructorProperties({"array"})
    public Matrix(double[][] array) {
        this.array = ArrayUtil.clone(array);
    }

    public Matrix(double[] array, int cols) {
        this.array = ArrayUtil.oneToTwo(array, cols);
    }

    public Matrix(Matrix matrix) {
        this.array = ArrayUtil.clone(matrix.array);
    }

    public Matrix(String path) {
        array = read(path);
    }

    /**
     * Ustawia wszystkie komórki macierzy na podaną wartość
     *
     * @param value - wartość do wstawienia
     */
    public void setToValue(double value) {
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                array[i][j] = value;
            }
        }
    }

    /**
     * @return zwraca liczbę rzędów w macierzy
     */
    public int getRowCount() {
        return array.length;
    }

    /**
     * @return zwraca liczbę kolumn w macierzy
     */
    public int getColCount() {
        return array[0].length;
    }

    /**
     * Drukuje tę macierz na konsoli
     */
    public void printToConsole() {
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                System.out.print(array[i][j] + " ");
            }
            System.out.println();
        }
    }

    public void printToConsole(int dec) {
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                System.out.print(MathUtil.roundToDecimal(array[i][j], dec)
                        + " ");
            }
            System.out.println();
        }
    }

    /**
     * Zapisuje macierz do podanego pliku. Wartości w rzędzie
     * oddzielone są spacją. Rząd kończy się średnikiem
     *
     * @param path - plik
     */
    public void write(String path) {
        try (PrintWriter pw = new PrintWriter(path)) {
            for (int i = 0; i < getRowCount(); i++) {
                for (int j = 0; j < getColCount() - 1; j++) {
                    pw.print(array[i][j] + " ");
                }
                pw.print(array[i][getColCount() - 1]);
                pw.print(";");
                pw.println();
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }

    /**
     * Wczytuje macierz z pliku
     *
     * @param path - ścieżka do pliku z macierzą
     * @return zwraca wczytaną macierz
     */
    public double[][] read(String path) {
        double[][] tabl;
        ArrayList<String> al = new ArrayList<>();
        BufferedReader in;
        StringTokenizer st;
        int len;
        int len1;
        try {
            in = new BufferedReader(new FileReader(new File(path)));
            try {
                String s;
                while ((s = in.readLine()) != null) {
                    al.add(s);
                }
            } finally {
                in.close();
            }
        } catch (IOException ioe) {
            throw new RuntimeException(ioe);
        }
        st = new StringTokenizer(al.get(0), " ;");
        len = st.countTokens();
        len1 = al.size();
        tabl = new double[len1][len];
        for (int i = 0; i < len1; i++) {
            st = new StringTokenizer(al.get(i), " ;");
            int j = 0;
            while (st.hasMoreTokens()) {
                tabl[i][j] = Double.parseDouble(st.nextToken());
                j++;
            }
        }
        return tabl;
    }

    /**
     * Zwraca głęboką kopię tej macierzy
     */
    @Override
    public Matrix clone() throws CloneNotSupportedException {
        Matrix cl = (Matrix) super.clone();
        cl.array = ArrayUtil.clone(array);
        return cl;
    }

    /**
     * Sprawdza czy macierz jest zerowa, tzn. czy wszystkie jej elementy
     * są równe 0.0
     *
     * @return true jeśli jest zerowa, false w przeciwnym wypadku
     */
    public boolean isZeroMatrix() {
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (array[i][j] != 0.0) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Tworzy macierz jednostkową. Na przekątnej są jedności, a
     * w pozostałych komórkach 0.
     *
     * @throws MatrixException - wyrzuca wyjątek
     */
    public void setToIdentity() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (i == j) {
                    array[i][j] = 1.0;
                } else {
                    array[i][j] = 0.0;
                }
            }
        }
    }

    /**
     * Sprawdza czy macierz jest jednostkowa. Macierz jest jednostkowa, gdy
     * wszystkie elementy macierzy są równe 0, z wyjątkiem elementów leżących na
     * przekątnej, które są równe 1.
     *
     * @return true jeśli macierz jest elementarna, false w przeciwnym
     * przypadku
     */
    public boolean isIdentityMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if ((i != j) && (array[i][j] != 0)) {
                    return false;
                } else if ((i == j) && (array[i][j] != 1)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest kwadratowa tzn. czy liczba rzędów
     * i kolumn jest jednakowa
     *
     * @return true jeśli jest kwadratowa, false w przeciwnym wypadku
     */
    public boolean isSquareMatrix() {
        return getRowCount() == getColCount();
    }

    /**
     * Sprawdza czy macierz jest wierszowa (jest wektorem wierszowym)
     *
     * @return true jeśli jest wierszowa, false w przeciwnym wypadku
     */
    public boolean isRowMatrix() {
        return getRowCount() == 1;
    }

    /**
     * Sprawdza czy macierz jest kolumnowa (jest wektorem kolumnowym)
     *
     * @return true jeśli jest kolumnowa, false w przeciwnym wypadku
     */
    public boolean isColumnMatrix() {
        return getColCount() == 1;
    }

    /**
     * Sprawdza czy macierz jest diagonalna. Macierz jest diagonalna jeśli
     * wszystkie elementy są zerowe z wyjątkiem elementów leżących na głównej
     * przekątnej
     *
     * @return true jeśli jest diagonalna, false w przeciwnym wypadku
     */
    public boolean isDiagonalMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if ((array[i][j] != 0.0) && (i != j)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Pobiera elementy leżące na przekątnej macierzy diagonalnej
     *
     * @return - tablicę elementów leżących na przekątnej
     */
    public double[] getDiagonal() {
        int len = Math.min(getRowCount(), getColCount());
        double[] diagonals = new double[len];
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (i == j) {
                    diagonals[i] = getArray()[i][j];
                }
            }
        }
        return diagonals;
    }

    public double[][] getArray() {
        return ArrayUtil.clone(array);
    }

    public void setArray(double[][] array) {
        this.array = ArrayUtil.clone(array);
    }

    /**
     * Sprawdza czy macierz jest macierzą skalarną
     *
     * @return - zwraca true jeśli macierz jest skalarna, false w
     * przeciwnym przypadku
     * @throws MatrixException - wyrzuca wyjątek
     */
    public boolean isScalarMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        double[] diagonals = getDiagonal();
        double first = diagonals[0];
        for (int i = 1; i < diagonals.length; i++) {
            if (diagonals[i] != first) {
                return false;
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest symetryczna. Symetryczna macierz to taka, w
     * której elementy po obu stronach przekątnej głównej są równe
     *
     * @return true jeśli jest symetryczna, false jeśli nie jest
     * symetryczna
     * @throws MatrixException - wyrzuca wyjątek
     */
    public boolean isSymmetricMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (array[i][j] != array[j][i]) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest antysymetryczna. Antysymetryczna to taka
     * macierz,w której elementy symetrycznie rozłożone po obu stronach głównej
     * przekątnej mają jednakowe wartości, ale z przeciwnym znakiem,
     * np. 3 i -3. Elementy diagonalne muszą być równe 0
     *
     * @return - true jeśli macierz jest antysymetryczna,
     * false w przeciwnym przypadku
     * @throws MatrixException - zwraca wyjątek
     */
    public boolean isAntisymmetricMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (array[i][j] != -array[j][i]) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest trójdiagonalna. W macierzy trójdiagonalnej
     * niezerowe mogą być tylko elementy leżące na głównej przekątnej i dwóch
     * przekątnych z nią sąsiadujących.
     *
     * @return true jeśli macierz jest trójdiagonalna, false
     * w przeciwnym wypadku
     * @throws MatrixException - wyrzucany wyjątek
     */
    public boolean is3DiagonalMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if ((array[i][j] == 0.0) && !(Math.abs(i - j) > 1)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest górnotrójkątna. Macierz jest
     * górnotrójkątna wtedy, gdy elementy  leżące poniżej głównej
     * przekątnej sa równe 0.
     *
     * @return true jeśli macierz jest górnotrójkątna, false
     * w przeciwnym wypadku
     * @throws MatrixException - wyrzucany wyjątek
     */
    public boolean isUpperTriangleMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if ((i > j) && (array[i][j] != 0)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest dolnotrójkątna. Macierz jest dolnotrójkątna,
     * gdy wszystkie elementy leżące powyżej głównej przekątnej są równe 0;
     *
     * @return true - jeśli macierz jest dolnotrójkątna, false w przeciwnym przypadku
     * @throws MatrixException - wyrzucany wyjątek
     */
    public boolean isLowerTriangleMatrix() throws MatrixException {
        if (!isSquareMatrix()) {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if ((i < j) && (array[i][j] != 0)) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Sprawdza czy macierz jest osobliwa
     *
     * @return - true jeśli macierz jest osobliwa, false w przeciwnym
     * przypadku
     */
    public boolean isSingularMatrix() {
        double det = 0.0;
        try {
            det = det();
        } catch (MatrixException e) {
            e.printStackTrace();
        }
        return det == 0;
    }

    public double det() throws MatrixException {
        double det = 0.0;
        if (isSquareMatrix()) {
            switch (getRowCount()) {
                case 1:
                    det = array[0][0];
                    break;
                case 2:
                    det = array[0][0] * array[1][1] - array[0][1] * array[1][0];
                    break;
                case 3:
                    double a = array[0][0];
                    double b = array[0][1];
                    double e = array[0][2];
                    double c = array[1][0];
                    double d = array[1][1];
                    double f = array[1][2];
                    double g = array[2][0];
                    double h = array[2][1];
                    double i = array[2][2];
                    det = ((a * d * i + b * f * g + e * c * h) - (e * d * g + a
                            * f * h + b * c * i));

                default:
                    int row = 0;
                    for (int j = 0; j < this.getColCount(); j++) {
                        try {
                            det += MatrixUtil.algComplement(this, row, j)
                                    * this.getCell(row, j);
                        } catch (MatrixException e1) {
                            e1.printStackTrace();
                        }
                    }
            }
        } else {
            throw new MatrixException("Macierz nie jest kwadratowa");
        }
        return det;
    }

    /**
     * Pobiera wartość umieszczoną w komórce
     *
     * @param row - rzad
     * @param col - kolumna
     * @return - wartość na przecięciu rzędu i kolumny
     */
    public double getCell(int row, int col) throws MatrixException {
        double v;
        if (isIndexValid(row, col)) {
            v = array[row][col];
        } else {
            throw new MatrixException("indeks rzędu lub kolumny poza macierzą");
        }
        return v;
    }

    /**
     * Zmienia wartość komórki na podaną
     *
     * @param row   - rząd
     * @param col   - kolumna
     * @param value - wartość do wstawienia
     */
    public void setCell(int row, int col, double value) throws MatrixException {
        if (isIndexValid(row, col)) {
            array[row][col] = value;
        } else {
            throw new MatrixException("indeks rzędu lub kolumny poza macierzą");
        }
    }

    /**
     * Sprawdza czy indeks kolumny mieści się w macierzy
     *
     * @param index - indeks do zbadania
     * @return zwraca true jeśli tal, false w przeciwnym wypadku
     */
    public boolean isColIndexValid(int index) {
        return index >= 0 && index <= (getColCount() - 1);
    }

    /**
     * Sprawdza czy indeks rzędu mieści się w macierzy
     *
     * @param index - indeks do zbadania
     * @return zwraca true jeśli index jest w macierzy
     * false w przeciwnym wypadku
     */
    public boolean isRowIndexValid(int index) {
        return index >= 0 && index <= (getRowCount() - 1);
    }

    /**
     * Sprawdza czy indeksy kolumny i rzędu są w obrębie macierzy.
     *
     * @param row - rząd
     * @param col - kolumna
     * @return zwraca true jeśli index jest w obrębie macierzy, false
     * w przeciwnym wypadku
     */
    public boolean isIndexValid(int row, int col) {
        return isRowIndexValid(row) && (isColIndexValid(col));
    }

    /**
     * Pobiera dane z kolumny
     *
     * @param index - index kolumny
     * @return - zwraca tablicę danych lub null jeśli index jest poza
     * macierzą
     */
    public double[] getCol(int index) {
        double[] tabl = null;
        if (isColIndexValid(index)) {
            tabl = new double[getRowCount()];
            for (int i = 0; i < tabl.length; i++) {
                tabl[i] = array[i][index];
            }
        }
        return tabl;
    }

    /**
     * Ustawia wartości kolumny na wartości z podanej tablicy
     *
     * @param index - indeks kolumny
     * @param col   - wartości do wstawienia
     * @throws MatrixException - wyrzucany wyjątek
     */
    public void setCol(int index, double[] col) throws MatrixException {
        if (isColIndexValid(index)) {
            for (int i = 0; i < col.length; i++) {
                array[i][index] = col[i];
            }
        } else {
            throw new MatrixException("indeks kolumny poza macierzą");
        }
    }

    /**
     * Zwraca dane z rzędu macierzy
     *
     * @param index - indeks rzędu
     * @return - pobrane dane lub null jeśli indeks rzędu jest poza macierzą
     */
    public double[] getRow(int index) {
        double[] tabl = null;
        if (isRowIndexValid(index)) {
            tabl = new double[getColCount()];
            System.arraycopy(array[index], 0, tabl, 0, tabl.length);
        }
        return tabl;
    }

    /**
     * Zmienia wartości w rzędzie na podane
     *
     * @param index - indeks rzędu
     * @param row   - dane do wstawienia
     * @throws MatrixException - wyrzucany wyjątek
     */
    public void setRow(int index, double[] row) throws MatrixException {
        if (isRowIndexValid(index)) {
            System.arraycopy(row, 0, array[index], 0, row.length);
        } else {
            throw new MatrixException("index rzędu poza macierzą");
        }
    }

    /**
     * Mnoży tę macierz przez podaną macierz i zwraca nową macierz
     *
     * @param n - macierz przez którą należy pomnożyć tę macierz
     * @return - nowa macierz będąca wynikiem pomnożenia tej macierzy przez
     * podaną macierz
     */
    public Matrix multiply2(Matrix n) {
        double[][] temp = new double[getRowCount()][n.getColCount()];
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < n.getColCount(); j++) {
                for (int k = 0; k < getColCount(); k++) {
                    try {
                        temp[i][j] += getCell(i, k) * n.getCell(k, j);
                    } catch (MatrixException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        return new Matrix(temp);
    }

    public boolean is3x3() {
        int r = getRowCount();
        int c = getColCount();
        return (r == 3) && (c == 3);
    }

    //wymiar (rozmiar, stopień) macierzy
    //to nie jest rząd macierzy
    public int[] size() {
        return new int[]{this.getRowCount(), getColCount()};
    }

    @Override
    public boolean equals(Object other) {
        if (this == other) {
            return false;
        }
        if (other == null) {
            return false;
        }
        if (getClass() != other.getClass()) {
            return false;
        }
        Matrix o = (Matrix) other;
        if (getRowCount() != o.getRowCount() || getColCount() != o.getColCount()) {
            return false;
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                if (array[i][j] != o.array[i][j]) {
                    return false;
                }
            }
        }
        return true;
    }

    @Override
    public int hashCode() {
        long b = 1L;
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                double cell = 0.0;
                try {
                    cell = getCell(i, j);
                } catch (MatrixException e) {
                    e.printStackTrace();
                }
                if (cell != 0.0) {
                    b = 31L * b + Double.doubleToLongBits(cell);
                } else {
                    b *= 31L;
                }
            }
        }
        return (int) (b ^ (b >> 32));
    }

    //Oblicza i zwraca ślad macierzy
    public double trace() {
        double[] arr = getDiagonal();
        double sum = 0.0;
        for (double v : arr) {
            sum += v;
        }
        return sum;
    }

    /**
     * Dodaje wskazana macierz do tej macierzy
     *
     * @param m - macierz do dodania
     * @throws MatrixException - wyrzucany wyjątek
     */
    public void add(Matrix m) throws MatrixException {
        if (hasSizeEqual(m)) {
            throw new MatrixException("macierze nierównej wielkości");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                array[i][j] += m.getCell(i, j);
            }
        }
    }

    /**
     * Zwraca nową macierz będącą wynikiem dodania wskazanej
     * macierzy do tej macierzy
     *
     * @param m - macierz do odjęcia
     * @throws MatrixException - wyrzucany wyjątek
     */
    public Matrix add2(Matrix m) throws MatrixException {
        if (hasSizeEqual(m)) {
            throw new MatrixException("macierze nierównej wielkości");
        }
        Matrix n = null;
        try {
            n = this.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        n.add(m);
        return n;
    }

    /**
     * Sprawdza czy macierze mają jednakową wielkość
     *
     * @param m - macierz do porównania
     * @return zwraca true jeśli macierze mają tę samą
     * wielkość, false - w przeciwnym wypadku
     */
    public boolean hasSizeEqual(Matrix m) {
        return (this.getRowCount() != m.getRowCount())
                || (this.getColCount() != m.getColCount());
    }

    /**
     * Odejmuje wskazaną macierz od tej macierzy
     *
     * @param m - macierz do odjęcia
     * @throws MatrixException - wyrzucany wyjątek
     */
    public void sub(Matrix m) throws MatrixException {
        if (hasSizeEqual(m)) {
            throw new MatrixException("macierze nierównej wielkości");
        }
        for (int i = 0; i < getRowCount(); i++) {
            for (int j = 0; j < getColCount(); j++) {
                array[i][j] -= m.getCell(i, j);
            }
        }
    }

    /**
     * Zwraca nowa macierz będąca wynikiem odjęcia wskazanej
     * macierzy od tej macierzy
     *
     * @param m -macierz do odjęcia
     * @throws MatrixException - wyrzucany wyjątek
     */
    public Matrix sub2(Matrix m) throws MatrixException {
        if (hasSizeEqual(m)) {
            throw new MatrixException("macierze nierównej wielkości");
        }
        Matrix n = null;
        try {
            n = this.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        n.sub(m);
        return n;
    }

}
